// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include "dump.h"
#include <i.h>
#include <math.h>

double fish_stark(struct device *in)
{
	struct istruct E;
	struct istruct phi;
	inter_import_array(&E,in->ymesh, in->phi,in->ymeshpoints,TRUE);
	inter_import_array(&phi,in->ymesh, in->phi,in->ymeshpoints,TRUE);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"fish_phi.dat");
	inter_deriv(&E,&phi);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"fish_E.dat");
	inter_pow(&E,2.0);
	inter_mul(&E,in->fish_stark);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"fish_x_stark.dat");
	double sum=inter_avg(&E);
	inter_free(&E);
	inter_free(&phi);

return sum;
}

void get_avg_np_pos(struct device *in,double *nx,double *px)
{
int i;
double navg=0.0;
double pavg=0.0;
double nsum=0.0;
double psum=0.0;


for (i=0;i<in->ymeshpoints;i++)
{
	navg+=(in->n[i]+in->nt_all[i]-in->n_orig[i])*in->ymesh[i];
	pavg+=(in->p[i]+in->pt_all[i]-in->p_orig[i])*in->ymesh[i];
	nsum+=(in->n[i]+in->nt_all[i]-in->n_orig[i]);
	psum+=in->p[i]+in->pt_all[i]-in->p_orig[i];

}
if (nsum!=0.0)
{
*nx=navg/nsum;
}else
{
*nx=0.0;
}

if (psum!=0.0)
{
*px=pavg/psum;
}else
{
*px=0.0;
}



}

void sim_fish(struct device *in)
{
int i;
double nx=0.0;
double px=0.0;
in->dt=in->fish_dt;

carrier_count_reset(in);

in->go_time=FALSE;

printf("Running fish. %e\n",in->fish_externalv);
int ittr=0;
int pittr=-1;
int step=0;

solve_light(in,&(in->mylight),in->Psun,0.0);

newton_sim_fish(in);

double stark_orig=fish_stark(in);

if (fabs(in->fish_externalv-in->Vapplied)>0.2)
{
	ramp(in,0.0,in->fish_externalv,7.0);
}

sim_externalv(in,in->fish_externalv);

newton_sim_fish(in);

time_init(in);

for (i=0;i<in->ymeshpoints;i++)
{
in->n_orig[i]=in->n[i]+in->nt_all[i];
in->p_orig[i]=in->p[i]+in->pt_all[i];

in->n_orig_f[i]=in->n[i];
in->p_orig_f[i]=in->p[i];

in->n_orig_t[i]=in->nt_all[i];
in->p_orig_t[i]=in->pt_all[i];
}




device_timestep(in);

in->go_time=TRUE;

in->time=in->fish_tstart;

	FILE *out=fopen("./fish.dat","w");
	fclose(out);

struct istruct fishoutn;
inter_init(&fishoutn);

struct istruct fishoutp;
inter_init(&fishoutp);

struct istruct fishoutj;
inter_init(&fishoutj);

struct istruct fishoutstark;
inter_init(&fishoutstark);

dump_dynamic_init(in);


printf("Main fish loop\n");
double stark=0.0;
do
{


newton_sim_fish(in);

//(in->time+in->dt>0.0)&&(in->time<=0.0)
if (in->time>=0.0)
{
	pittr++;
}

if (pittr==0)
{
	for (i=0;i<in->ymeshpoints;i++)
	{
		if ((in->ymesh[i]>in->fish_1start)&&(in->ymesh[i]<in->fish_1stop))
		{
			in->p[i]+=in->fish_den;
		}

		if ((in->ymesh[i]>in->fish_2start)&&(in->ymesh[i]<in->fish_2stop))
		{
			in->n[i]+=in->fish_den;
		}
		if (get_dump_status(dump_print_text)==TRUE)
		{
			printf("%d %le %le %le\n",i,in->ymesh[i],in->n[i],in->p[i]);
		}
	}
}

	dump_dynamic_add_data(in);

	stark=fish_stark(in);



	ittr++;

	dump_slice(in);
	plot_now(in,"fish.plot");
	if (in->time>=0.0)
	{
	get_avg_np_pos(in,&nx,&px);
	inter_append(&fishoutn,in->time,nx);
	inter_append(&fishoutp,in->time,px);
	}
	inter_append(&fishoutstark,in->time,stark);

	//getchar();
	inter_append(&fishoutj,in->time,get_J(in));

	//if (in->time>5e-11) in->stop_start=TRUE;

//if (in->time>0.0) getchar();
	if (get_dump_status(dump_print_text)==TRUE)
	{
		printf("fish step t=%le V=%le current=%le %le\n",in->time,in->Vapplied,get_I(in),in->last_error);
	}

	device_timestep(in);




	stop_start(in);

	if (in->time>0.0) in->dt*=in->fish_dt_mull;

	step++;

	//printf("%lf\n",in->Vapplied);
	//FILE *out=fopen("./fish.dat","a");
	//fprintf(out,"%le %le %le\n",in->time,nx,px);
	//fclose(out);

	if (in->time>in->fish_stop) break;

	if (in->stop==TRUE)
	{
		break;
	}


}while(1);

dump_dynamic_save(in);
dump_dynamic_free(in);

in->go_time=FALSE;

char outpath[1000];
sprintf(outpath,"%s%s",in->outputpath,"fishn.dat");
inter_sub(&fishoutn,fishoutn.data[0]);
inter_mod(&fishoutn);
inter_sub(&fishoutp,fishoutp.data[0]);
inter_mod(&fishoutp);
if (get_dump_status(dump_iodump)==TRUE) inter_save(&fishoutn,outpath);



sprintf(outpath,"%s%s",in->outputpath,"fishp.dat");
if (get_dump_status(dump_iodump)==TRUE) inter_save(&fishoutp,outpath);

inter_add(&fishoutn,&fishoutp);

sprintf(outpath,"%s%s",in->outputpath,"fishnp.dat");
inter_save(&fishoutn,outpath);

struct istruct v;
inter_copy(&v,&fishoutn,TRUE);
inter_deriv(&v,&fishoutn);
if (get_dump_status(dump_iodump)==TRUE) inter_save(&v,"fish_driv.dat");
double E=fabs(in->Vapplied)/in->ymesh[in->ymeshpoints-1];
for (i=0;i<fishoutn.len;i++)
{
	fishoutn.data[i]=(v.data[i]/E);
}
sprintf(outpath,"%s%s",in->outputpath,"fish_mu.dat");
if (get_dump_status(dump_iodump)==TRUE) inter_save(&fishoutn,outpath);
int mid=in->ymeshpoints/2;
double xl=in->ymesh[mid-1];
double xr=in->ymesh[mid+1];
double phil=in->phi[mid-1];
double phir=in->phi[mid+1];
E=fabs((phir-phil)/(xr-xl));
for (i=0;i<fishoutn.len;i++)
{
	fishoutn.data[i]=(v.data[i]/E);
}
sprintf(outpath,"%s%s",in->outputpath,"fish_mu_local_E.dat");
if (get_dump_status(dump_iodump)==TRUE) inter_save(&fishoutn,outpath);


sprintf(outpath,"%s%s",in->outputpath,"fishj.dat");
if (get_dump_status(dump_iodump)==TRUE) inter_save(&fishoutj,outpath);


sprintf(outpath,"%s%s",in->outputpath,"fish_stark.dat");
inter_save(&fishoutstark,outpath);

//struct istruct file;
//inter_load_a(&file,"fish_orig.dat",0,1);
//for (i=0;i<file.len;i++)
//{
//fishoutstark.data[i]-=file.data[i];
//}

//sprintf(outpath,"%s%s",in->outputpath,"fish_stark_sub.dat");
//inter_save(&fishoutstark,outpath);

//inter_free(&file);


inter_free(&v);
inter_free(&fishoutn);
inter_free(&fishoutp);
inter_free(&fishoutj);
inter_free(&fishoutstark);
}
