// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <time.h>
#include <dirent.h>
#include <util.h>
#include <pwd.h>
#include <unistd.h>
#include "lock.h"
#include "sim.h"
#include "dos.h"
#include "server.h"
#include "fit.h"
#include "sun.h"
#include "dump.h"
#include "complex_solver.h"


struct device cell;

void printdebug(char *text)
{

printf("%s",text);

}

void dialog_set_progress (double percent)
{

}

void device_init(struct device *in)
{
	in->Voc=0.0;
	in->Jsc=0.0;
	in->FF=0.0;
	in->Pmax=0.0;
	in->Pmax_voltage=0.0;
	device_get_memory(in);

}

void find_n0(struct device *in)
{
printf("Finding n0\n");
double oldsun=in->Psun;
double oldv=in->Vapplied;
double B_temp;
in->Vapplied=0;
in->Psun=0;

solve_light(in,&(in->mylight),in->Psun,0.0);

dump_for_plot(in);

B_temp=in->B;
in->B=0.0;
double save_clamp=in->electrical_clamp;
int save_ittr=in->max_electrical_itt;
in->electrical_clamp=in->electrical_clamp0;
in->max_electrical_itt=in->max_electrical_itt0;
solve_all(in);
in->max_electrical_itt=save_ittr;
in->electrical_clamp=save_clamp;
solve_all(in);
in->B=B_temp;
//getchar();
reset_npinit(in);
reset_npequlib(in);

in->Psun=oldsun;
in->Vapplied=oldv;
solve_light(in,&(in->mylight),in->Psun,0.0);

}



void quick_update()
{
}



int run_simulation(char *runpath)
{
printf("Run_simulation\n");
if (strcmp(runpath,"")!=0) strcpy(cell.outputpath,runpath);
int i;

cell.kTq=(300.0*kb/q);
printf("load config\n");
load_config("device.inp",&cell,FALSE);

printf("dos init for %d layers\n",cell.mat_layers);
char tempn[100];
char tempp[100];
for (i=0;i<cell.mat_layers;i++)
{
	dos_init(i);
	printf("load dos %d/%d\n",i,cell.mat_layers);
	sprintf(tempn,"dosn%d.dat",i);
	sprintf(tempp,"dosp%d.dat",i);
	load_dos(&cell,tempn,tempp,i);
}

device_init(&cell);

#ifdef dump_converge
cell.converge = fopena(cell.outputpath,"./converge.dat","w");
fclose(cell.converge);

cell.tconverge=fopena(cell.outputpath,"./tconverge.dat","w");
fclose(cell.tconverge);
#endif


int curlayer=0;
double end=cell.mat.l[0].height;

for (i=0;i<cell.ymeshpoints;i++)
{
	if (cell.ymesh[i]>=end)
	{
		curlayer++;
		end+=cell.mat.l[curlayer].height;
	}
	cell.imat[i]=curlayer;
	cell.Nad[i]=cell.mat.l[cell.imat[i]].Nad+cell.doping;

}
//getchar();

	for (i=0;i<cell.ymeshpoints;i++)
	{
		cell.Tl[i]=cell.Tll+cell.ymesh[i]*(cell.Tlr-cell.Tll)/cell.ylen;
		cell.Te[i]=cell.Tll+cell.ymesh[i]*(cell.Tlr-cell.Tll)/cell.ylen;
		cell.Th[i]=cell.Tll+cell.ymesh[i]*(cell.Tlr-cell.Tll)/cell.ylen;
		cell.ex[i]=0.0;
		cell.Hex[i]=0.0;
		//if ((i>cell.ymeshpoints/2)&&(i<cell.ymeshpoints/2+10)) cell.Hex[i]=1e9;
		cell.epsilonr[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_epsilonr);
		cell.dostype[i]=(int)get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_dostype);

		cell.Eg[i]=get_dos_Eg(cell.imat[i]);
		cell.Dex[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_Dex);

		cell.Xi[i]=get_dos_Xi(cell.imat[i]);

		cell.Ec[i]=-cell.Xi[i];

		cell.Ev[i]=-cell.Xi[i]-cell.Eg[i];

		//printf("%d %e %e\n",i,cell.mun[i],cell.mup[i]);

		cell.Nc[i]=get_Nc_free(cell.imat[i]);

		cell.Nv[i]=get_Nv_free(cell.imat[i]);

		cell.mun[i]=get_n_mu(cell.imat[i]);
		cell.mup[i]=get_p_mu(cell.imat[i]);

		cell.kf[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_kf);
		cell.a[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_a);
		cell.kl[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_kl);
		cell.ke[i]=get_n_mu(cell.imat[i]);
		cell.kh[i]=get_p_mu(cell.imat[i]);

		cell.Hl[i]=0.0;
		cell.He[i]=0.0;
		cell.Hh[i]=0.0;
		cell.Habs[i]=0.0;

		//Electron capture
		//getchar();


		//printf("%e %e %e %e\n",cell.taue0,cell.taue1,cell.tauh0,cell.tauh1);
		//getchar();
		cell.deltaE[i]=get_mat_param(&(cell.mat.l[cell.imat[i]]),mat_deltaE);
		cell.capeff[i]=get_mat_param_poly(&(cell.mat.l[cell.imat[i]]),mat_capeff,0.0);
	}
//getchar();
	for (i=0;i<cell.ymeshpoints;i++)
	{
		cell.phi[i]=0.0;
		cell.R[i]=0.0;
		cell.n[i]=0.0;
	}

cell.C=cell.xlen*cell.zlen*epsilon0*cell.epsilonr[0]/(cell.ylen+cell.other_layers);
if (get_dump_status(dump_print_text)==TRUE) printf("C=%le\n",cell.C);
cell.A=cell.xlen*cell.zlen;
cell.Vol=cell.xlen*cell.zlen*cell.ylen;

light_set_dx(&cell.mylight,cell.ymesh[1]-cell.ymesh[0]);
light_load_config(&cell.mylight,cell.outputpath);
if (get_dump_status(dump_iodump)==FALSE) set_dump_status(dump_optics,FALSE);

//update_arrays(&cell);

cell.Vapplied=0.0;
get_initial(&cell);
solve_pos(&cell);

time_init(&cell);

cell.N=0;
cell.M=0;

solver_realloc(&cell);

if (cell.onlypos==TRUE)
{
	device_free(&cell);
	return 0;
}

plot_open(&cell);

cell.stop=FALSE;
cell.go_time=FALSE;

plot_now(&cell,"plot");

tau_test(&cell);

find_n0(&cell);


dirty_fit(&cell);

draw_gaus(&cell);


if (cell.simmode==sim_mode_jv)
{
	if (lock_allow_jv()==TRUE) 	sim_jv(&cell);
}
else
if (cell.simmode==sim_mode_tpc)
{
	if (lock_allow_tpc()==TRUE) sim_tpc(&cell);
}else
if (cell.simmode==sim_mode_ce)
{
	if (lock_allow_ce()==TRUE) sim_ce(&cell);
}
else
if (cell.simmode==sim_mode_li)
{
	if (lock_allow_li()==TRUE) sim_li(&cell);
}else
if (cell.simmode==sim_mode_li_sc)
{
	if (lock_allow_li()==TRUE) sim_li_sc(&cell);
}else
if (cell.simmode==sim_mode_tpv)
{
	if (lock_allow_tpv()==TRUE) sim_tpv(&cell);
}else
if (cell.simmode==sim_mode_findvoc)
{
	if (lock_allow_voc()==TRUE) sim_voc(&cell);
}else
if (cell.simmode==sim_mode_findlight_for_voc)
{
	if (lock_allow_light_for_voc()==TRUE) set_light_for_voc(&cell,cell.tpvvoc);
}
else
if (cell.simmode==sim_mode_jvone)
{
	if (lock_allow_jv_one_point()==TRUE) sim_jv_one_point(&cell);
}else
if (cell.simmode==sim_mode_tof)
{
	if (lock_allow_tof()==TRUE) sim_tof(&cell);
}else
if (cell.simmode==sim_mode_celiv)
{
	if (lock_allow_celiv()==TRUE) sim_celiv(&cell);
}else
if (cell.simmode==sim_mode_fish)
{
	if (lock_allow_fish()==TRUE) sim_fish(&cell);
}else
if (cell.simmode==sim_mode_simple_jv)
{
//	if (lock_allow_fish()==TRUE)
	sim_simple_jv(&cell);
}else
if (cell.simmode==sim_mode_qe)
{
//	if (lock_allow_fish()==TRUE)
	sim_qe(&cell);
}else
if (cell.simmode==sim_mode_pulse)
{
//	if (lock_allow_fish()==TRUE)
	sim_pulse(&cell);
}



device_free(&cell);

plot_close(&cell);

for (i=0;i<cell.mat_layers;i++)
{
	dos_free(i);
}

solver_free_memory(&cell);
light_free(&cell.mylight);

return cell.odes;
}

struct device gui_device;




int main (int argc, char *argv[])
{
if ((scanarg( argv,argc,"--fit")==TRUE)||(scanarg( argv,argc,"--1fit")==TRUE))
{
set_dump_status(dump_lock, FALSE);
set_dump_status(dump_print_text, FALSE);
set_dump_status(dump_iodump, FALSE);
set_dump_status(dump_optics, FALSE);
set_dump_status(dump_newton, FALSE);
set_dump_status(dump_plot, FALSE);
set_dump_status(dump_stop_plot, FALSE);
set_dump_status(dump_opt_for_fit, FALSE);
set_dump_status(dump_lock, TRUE);
}
if(geteuid()==0) {
	printf("Don't run me as root!\n");
	exit(0);
}
set_dump_status(dump_stop_plot, FALSE);
set_dump_status(dump_print_text, TRUE);
lock_int3
set_io_dump(FALSE);
srand(time(0));
textcolor(fg_green);
randomprint("Organic Photovoltaic Device Model (www.opvdm.com)\n");
randomprint("Copyright Roderick MacKenzie 2012 all rights reserved\n");
randomprint("\n");
randomprint("If you wish to collaborate in anyway please get in touch:\n");
randomprint("roderick.mackenzie@nottingham.ac.uk\n");
randomprint("www.roderickmackenzie.eu/contact.html\n");
//time_t t;
//printf("At begining %ld\n",time(NULL));
//for (t = time(NULL) + 5; time(NULL) < t; ) {}
//printf("At begining %ld\n",time(NULL));

char lock_path[100];
struct passwd *pw = getpwuid(getuid());
const char *homedir = pw->pw_dir;
FILE *in;
#ifndef nolock
sprintf(lock_path,"%s/.opvdm_config",homedir);

in = fopen(lock_path, "rb");

if (in==NULL)
{
exit(0);
}
fclose(in);
#endif

randomprint("\n");
textcolor(fg_reset);
globalserver.on=FALSE;
globalserver.cpus=1;
globalserver.readconfig=TRUE;
cell.fast=FALSE;
cell.newton_cur=FALSE;
//printf("%le\n",read_value(0,2));
//getchar();
//copy_file("",2,1e-2);
//exit(0);
//solve_bi();

double a;
double b;
double c;
double d;

if (scanarg( argv,argc,"--outputpath")==TRUE)
{
	strcpy(cell.outputpath,get_arg_plusone( argv,argc,"--outputpath"));
}else
{
	strcpy(cell.outputpath,"./");
}

if (scanarg( argv,argc,"--cpus")==TRUE)
{
sscanf(get_arg_plusone( argv,argc,"--cpus"),"%d",&globalserver.cpus);
globalserver.readconfig=FALSE;
}

if (scanarg( argv,argc,"--optics")==TRUE)
{
struct light two;
light_load_config(&two,"optics.inp");
solve_light(&cell,&two,1000.0,0.0);
light_free(&two);
return 0;
}

if (scanarg( argv,argc,"--norm")==TRUE)
{
struct istruct file;
double my_pos;
inter_load_a(&file,get_arg_plusone( argv,argc,"--norm"),0,1);
sscanf(get_arg_plusone( argv,argc,"--norm_pos"),"%lf",&my_pos);
double val=inter_get_hard(&file,my_pos);

inter_rescale(&file,1.0, 1.0/val);
inter_save_a(&file,cell.outputpath,get_arg_plusone( argv,argc,"--norm"));
inter_free(&file);
exit(0);
}else
if (scanarg( argv,argc,"--tab")==TRUE)
{
dump_talbe();
exit(0);
}else
if (scanarg( argv,argc,"--processce")==TRUE)
{
int col;
sscanf(get_arg_plusone( argv,argc,"--col"),"%d",&col);
process_ce_data(col,get_arg_plusone( argv,argc,"--in"),get_arg_plusone( argv,argc,"--out"));
exit(0);
}else
if (scanarg( argv,argc,"--qe")==TRUE)
{
dump_qe();
exit(0);
}
else
if (scanarg( argv,argc,"--fittpv")==TRUE)
{
double start;
double stop;
sscanf(argv[3],"%le",&start);
sscanf(argv[4],"%le",&stop);
sscanf(argv[5],"%le",&a);
sscanf(argv[6],"%le",&b);
sscanf(argv[7],"%le",&c);


fit_tpv(argv[2],start,stop,a,b,c);
exit(0);
}else
if (scanarg( argv,argc,"--fittpc")==TRUE)
{
//8.00834e-08,1.66422e-07,9.02609e-07,4.34910e-06

sscanf(argv[3],"%le",&a);
sscanf(argv[4],"%le",&b);
sscanf(argv[5],"%le",&c);
sscanf(argv[6],"%le",&d);

get_tpc_t0(argv[2],a,b,c,d);
exit(0);
}
else
if (scanarg( argv,argc,"--jointpvandce")==TRUE)
{
load_config("device.inp",&cell,FALSE);
join_tpv_to_ce(&cell);
exit(0);
}
else
if (scanarg( argv,argc,"--fft")==TRUE)
{
	double T=300.0;
	double E_start=0.0;
	double E_stop=0.15;
	double t_start=1e-9;
	double t_stop=0.01;
	int do_fit=FALSE;
	double norm_E_pos=0.2;
	if (scanarg( argv,argc,"--norm_E_pos")==TRUE) sscanf(get_arg_plusone( argv,argc,"--norm_E_pos"),"%lf",&norm_E_pos);
	if (scanarg( argv,argc,"--T")==TRUE) sscanf(get_arg_plusone( argv,argc,"--T"),"%lf",&T);
	if (scanarg( argv,argc,"--E_start")==TRUE) sscanf(get_arg_plusone( argv,argc,"--E_start"),"%lf",&E_start);
	if (scanarg( argv,argc,"--E_stop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--E_stop"),"%lf",&E_stop);
	if (scanarg( argv,argc,"--t_start")==TRUE) sscanf(get_arg_plusone( argv,argc,"--t_start"),"%lf",&t_start);
	if (scanarg( argv,argc,"--t_stop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--t_stop"),"%lf",&t_stop);
	if (scanarg( argv,argc,"--fftfit")==TRUE) do_fit=TRUE;


	fft_extract(cell.outputpath,get_arg_plusone( argv,argc,"--fft"),T,E_start,E_stop,t_start,t_stop,do_fit,norm_E_pos);
	exit(0);
}else
if (scanarg( argv,argc,"--pt")==TRUE)
{
	double T=300.0;
	double mull=1e-4;
	int ptfit=FALSE;
	double start=0.0;
	double stop=1.0;
	double dosstart=0.1;
	double dosstop=0.5;
	double chop=0.45;
	double chop1=0.0;
	double norm_point=0.01;
	double norm_E_pos=0.27;
	double sigmain=0.0001;
	double norm_E_mull=1.0;
	double v0=2e7;
	int maxittr=4;
	if (scanarg( argv,argc,"--ptfit")==TRUE) ptfit=TRUE;
	if (scanarg( argv,argc,"--norm_point")==TRUE) sscanf(get_arg_plusone( argv,argc,"--norm_point"),"%lf",&norm_point);
	if (scanarg( argv,argc,"--T")==TRUE) sscanf(get_arg_plusone( argv,argc,"--T"),"%lf",&T);
	if (scanarg( argv,argc,"--start")==TRUE) sscanf(get_arg_plusone( argv,argc,"--start"),"%lf",&start);
	if (scanarg( argv,argc,"--stop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--stop"),"%lf",&stop);
	if (scanarg( argv,argc,"--stop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--mull"),"%lf",&mull);
	if (scanarg( argv,argc,"--dosstart")==TRUE) sscanf(get_arg_plusone( argv,argc,"--dosstart"),"%lf",&dosstart);
	if (scanarg( argv,argc,"--dosstop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--dosstop"),"%lf",&dosstop);
	if (scanarg( argv,argc,"--maxittr")==TRUE) sscanf(get_arg_plusone( argv,argc,"--maxittr"),"%d",&maxittr);
	if (scanarg( argv,argc,"--chop")==TRUE) sscanf(get_arg_plusone( argv,argc,"--chop"),"%lf",&chop);
	if (scanarg( argv,argc,"--chop1")==TRUE) sscanf(get_arg_plusone( argv,argc,"--chop1"),"%lf",&chop1);
	if (scanarg( argv,argc,"--norm_E_pos")==TRUE) sscanf(get_arg_plusone( argv,argc,"--norm_E_pos"),"%lf",&norm_E_pos);
	if (scanarg( argv,argc,"--sigmain")==TRUE) sscanf(get_arg_plusone( argv,argc,"--sigmain"),"%lf",&sigmain);
	if (scanarg( argv,argc,"--v0")==TRUE) sscanf(get_arg_plusone( argv,argc,"--v0"),"%lf",&v0);
	if (scanarg( argv,argc,"--norm_E_mull")==TRUE) sscanf(get_arg_plusone( argv,argc,"--norm_E_mull"),"%lf",&norm_E_mull);

	pt_extract(cell.outputpath,get_arg_plusone( argv,argc,"--pt"),T,ptfit,start,stop,mull,dosstart,dosstop,maxittr,chop,chop1,norm_point,norm_E_pos,sigmain,v0,norm_E_mull);
	exit(0);
}
//printf("Running\n");
//getchar();
//run_simulation("./");
//exit(0);
char name[1000];
FILE *config=fopena(cell.outputpath,"ver.inp","r");
fscanf(config,"%s",name);
fclose(config);

if (strcmp(name,version)!=0)
{
printf("Software is version %s and the input files are version %s\n",version,name);
exit(0);
}

server_init(&globalserver);

if (scanarg( argv,argc,"--setup-network")==TRUE)
{
char user[255];
char ip[255];
lock_pw();
printf("user:");
fgets(user,20,stdin);
printf("IP:");
fgets(ip,20,stdin);
if(ip[strlen(ip) - 1] == '\n') ip[strlen(ip) - 1] = 0;
if(user[strlen(user) - 1] == '\n') user[strlen(user) - 1] = 0;

in = fopen(lock_path, "wb");
fclose(in);

lock_get_code(user,ip,TRUE);
exit(0);
}

if (scanarg( argv,argc,"--genkey")==TRUE)
{
char user[255];
printf("user:");
fgets(user,20,stdin);
if(user[strlen(user) - 1] == '\n') user[strlen(user) - 1] = 0;

in = fopen(lock_path, "wb");
fclose(in);

lock_get_code(user,"",FALSE);
printf("Your key has been written to key.dat send\nit to one of the above e-mail addresses\nto obtain a license.\n");
exit(0);
}
//printf("echo1 %ld\n",time(NULL));
square();
//printf("echo2 %ld\n",time(NULL));

if (scanarg( argv,argc,"--rights")==TRUE)
{
lock_print_license();
exit(0);
}else
if (scanarg( argv,argc,"--1fit")==TRUE)
{
lock_int3
lock_fit_enable();
static struct fitvars fitconfig;
set_dump_status(dump_lock, FALSE);
set_dump_status(dump_print_text, FALSE);
set_dump_status(dump_iodump, FALSE);
set_dump_status(dump_optics, FALSE);
set_dump_status(dump_newton, FALSE);
set_dump_status(dump_plot, FALSE);
set_dump_status(dump_stop_plot, FALSE);
set_dump_status(dump_opt_for_fit, FALSE);
set_dump_status(dump_lock, TRUE);
fit_read_config(&fitconfig);

fit_run_sims(&fitconfig,&globalserver);

}else
if (scanarg( argv,argc,"--fit")==TRUE)
{
lock_int3
lock_fit_enable();
FILE *out;
out = fopen("./fitlog.dat","w");
fclose(out);
int cont=TRUE;
	cell.fast=TRUE;
	do
	{
	cont=fit_now();

	}while(cont);



}else
{

	gen_dos_fd_gaus_fd();

	server_add_job(&globalserver,"./");
	print_jobs(&globalserver);
	server_run_jobs(&globalserver);
}
//printf("echo3 %ld\n",time(NULL));
server_shut_down(&globalserver);

return 0;
}

